@@ -20,9 +20,15 @@ typedef struct {
     char* notify_ext;
     /* kill all processes within a process group */
     bool kill_process_group;
+    /* do not kill processes owned by root */
+    bool ignore_root_user;
+    /* find process with the largest rss */
+    bool sort_by_rss;
     /* prefer/avoid killing these processes. NULL = no-op. */
     regex_t* prefer_regex;
     regex_t* avoid_regex;
+    /* will ignore these processes. NULL = no-op. */
+    regex_t* ignore_regex;
     /* memory report interval, in milliseconds */
     int report_interval_ms;
     /* Flag --dryrun was passed */
@@ -31,5 +37,6 @@ typedef struct {
 
 void kill_process(const poll_loop_args_t* args, int sig, const procinfo_t* victim);
 procinfo_t find_largest_process(const poll_loop_args_t* args);
+bool is_larger(const poll_loop_args_t* args, const procinfo_t* victim, procinfo_t* cur);
 
 #endif
