@@ -39,6 +39,10 @@ enum {
     LONG_OPT_PREFER = 513,
     LONG_OPT_AVOID,
     LONG_OPT_DRYRUN,
+    LONG_OPT_IGNORE,
+    LONG_OPT_IGNORE_ROOT,
+    LONG_OPT_USE_SYSLOG,
+    LONG_OPT_SORT_BY_RSS,
 };
 
 static int set_oom_score_adj(int);
@@ -57,12 +61,6 @@ double min(double x, double y)
     return y;
 }
 
-void handle_sigchld(int sig)
-{
-    (void)sig; // unused
-    waitpid(-1, NULL, WNOHANG);
-}
-
 // Dry-run oom kill to make sure that
 // (1) it works (meaning /proc is accessible)
 // (2) the stack grows to maximum size before calling mlockall()
@@ -81,6 +79,27 @@ static void startup_selftests(poll_loop_args_t* args)
             warn("%s: -N: notify script '%s' is not executable: %s\n", __func__, args->notify_ext, strerror(errno));
         }
     }
+
+#ifdef PROFILE_FIND_LARGEST_PROCESS
+    struct timespec t0 = { 0 }, t1 = { 0 };
+    clock_gettime(CLOCK_MONOTONIC, &t0);
+
+    warn("PROFILE_FIND_LARGEST_PROCESS: looping forever on find_largest_process(). Use sysprof of perf to capture profile.\n");
+    long i = 0;
+    while (1) {
+        find_largest_process(args);
+        i++;
+
+        const int avg_n = 1000;
+        if (i % avg_n == 0) {
+            clock_gettime(CLOCK_MONOTONIC, &t1);
+            long delta_usecs = (t1.tv_sec - t0.tv_sec) * 1000000 + (t1.tv_nsec - t0.tv_nsec) / 1000;
+            double avg_wall_time_ms = (double)(delta_usecs / avg_n) / 1000.0;
+            info("average find_largest_process() wall time: %.3lf ms\n", avg_wall_time_ms);
+            clock_gettime(CLOCK_MONOTONIC, &t0);
+        }
+    };
+#endif
 }
 
 int main(int argc, char* argv[])
@@ -91,20 +110,24 @@ int main(int argc, char* argv[])
         .mem_kill_percent = 5,
         .swap_kill_percent = 5,
         .report_interval_ms = 1000,
+        .ignore_root_user = false,
+        .sort_by_rss = false,
         /* omitted fields are set to zero */
     };
     int set_my_priority = 0;
     char* prefer_cmds = NULL;
     char* avoid_cmds = NULL;
+    char* ignore_cmds = NULL;
     regex_t _prefer_regex;
     regex_t _avoid_regex;
+    regex_t _ignore_regex;
 
     /* request line buffering for stdout - otherwise the output
      * may lag behind stderr */
     setlinebuf(stdout);
 
     /* clean up dbus-send zombies */
-    signal(SIGCHLD, handle_sigchld);
+    signal(SIGCHLD, SIG_IGN);
 
     fprintf(stderr, "earlyoom " VERSION "\n");
 
@@ -112,11 +135,14 @@ int main(int argc, char* argv[])
         fatal(4, "Could not cd to /proc: %s", strerror(errno));
     }
 
+    // PR_CAP_AMBIENT is not available on kernel < 4.3
+#ifdef PR_CAP_AMBIENT
     // When systemd starts a daemon with capabilities, it uses ambient
     // capabilities to do so. If not dropped, the capabilities can spread
     // to any child process. This is usually not necessary and its a good
     // idea to drop them if not needed.
     prctl(PR_CAP_AMBIENT, PR_CAP_AMBIENT_CLEAR_ALL, 0, 0, 0);
+#endif
 
     meminfo_t m = parse_meminfo();
 
@@ -125,8 +151,13 @@ int main(int argc, char* argv[])
     struct option long_opt[] = {
         { "prefer", required_argument, NULL, LONG_OPT_PREFER },
         { "avoid", required_argument, NULL, LONG_OPT_AVOID },
+        { "ignore", required_argument, NULL, LONG_OPT_IGNORE },
         { "dryrun", no_argument, NULL, LONG_OPT_DRYRUN },
+        { "ignore-root-user", no_argument, NULL, LONG_OPT_IGNORE_ROOT },
+        { "sort-by-rss", no_argument, NULL, LONG_OPT_SORT_BY_RSS },
+        { "syslog", no_argument, NULL, LONG_OPT_USE_SYSLOG },
         { "help", no_argument, NULL, 'h' },
+        { "debug", no_argument, NULL, 'd' },
         { 0, 0, NULL, 0 } /* end-of-array marker */
     };
     bool have_m = 0, have_M = 0, have_s = 0, have_S = 0;
@@ -214,6 +245,14 @@ int main(int argc, char* argv[])
         case 'p':
             set_my_priority = 1;
             break;
+        case LONG_OPT_IGNORE_ROOT:
+            args.ignore_root_user = true;
+            fprintf(stderr, "Processes owned by root will not be killed\n");
+            break;
+        case LONG_OPT_SORT_BY_RSS:
+            args.sort_by_rss = true;
+            fprintf(stderr, "Find process with the largest rss\n");
+            break;
         case LONG_OPT_PREFER:
             prefer_cmds = optarg;
             break;
@@ -224,6 +263,12 @@ int main(int argc, char* argv[])
             warn("dryrun mode enabled, will not kill anything\n");
             args.dryrun = 1;
             break;
+        case LONG_OPT_USE_SYSLOG:
+            earlyoom_syslog_init();
+            break;
+        case LONG_OPT_IGNORE:
+            ignore_cmds = optarg;
+            break;
         case 'h':
             fprintf(stderr,
                 "Usage: %s [OPTION]...\n"
@@ -241,15 +286,19 @@ int main(int argc, char* argv[])
                 "  -n                        enable d-bus notifications\n"
                 "  -N /PATH/TO/SCRIPT        call script after oom kill\n"
                 "  -g                        kill all processes within a process group\n"
-                "  -d                        enable debugging messages\n"
+                "  -d, --debug               enable debugging messages\n"
                 "  -v                        print version information and exit\n"
                 "  -r INTERVAL               memory report interval in seconds (default 1), set\n"
                 "                            to 0 to disable completely\n"
                 "  -p                        set niceness of earlyoom to -20 and oom_score_adj to\n"
                 "                            -100\n"
+                "  --ignore-root-user        do not kill processes owned by root\n"
+                "  --sort-by-rss             find process with the largest rss (default oom_score)\n"
                 "  --prefer REGEX            prefer to kill processes matching REGEX\n"
                 "  --avoid REGEX             avoid killing processes matching REGEX\n"
+                "  --ignore REGEX            ignore processes matching REGEX\n"
                 "  --dryrun                  dry run (do not kill any processes)\n"
+                "  --syslog                  use syslog instead of std streams\n"
                 "  -h, --help                this help text\n",
                 argv[0]);
             exit(0);
@@ -304,6 +353,13 @@ int main(int argc, char* argv[])
         }
         fprintf(stderr, "Will avoid killing process names that match regex '%s'\n", avoid_cmds);
     }
+    if (ignore_cmds) {
+        args.ignore_regex = &_ignore_regex;
+        if (regcomp(args.ignore_regex, ignore_cmds, REG_EXTENDED | REG_NOSUB) != 0) {
+            fatal(6, "could not compile regexp '%s'\n", ignore_cmds);
+        }
+        fprintf(stderr, "Will ignore process names that match regex '%s'\n", ignore_cmds);
+    }
     if (set_my_priority) {
         bool fail = 0;
         if (setpriority(PRIO_PROCESS, 0, -20) != 0) {
@@ -321,11 +377,11 @@ int main(int argc, char* argv[])
     }
 
     // Print memory limits
-    fprintf(stderr, "mem total: %4lld MiB, swap total: %4lld MiB\n",
-        m.MemTotalMiB, m.SwapTotalMiB);
-    fprintf(stderr, "sending SIGTERM when mem <= " PRIPCT " and swap <= " PRIPCT ",\n",
+    fprintf(stderr, "mem total: %4lld MiB, user mem total: %4lld MiB, swap total: %4lld MiB\n",
+        m.MemTotalKiB / 1024, m.UserMemTotalKiB / 1024, m.SwapTotalKiB / 1024);
+    fprintf(stderr, "sending SIGTERM when mem avail <= " PRIPCT " and swap free <= " PRIPCT ",\n",
         args.mem_term_percent, args.swap_term_percent);
-    fprintf(stderr, "        SIGKILL when mem <= " PRIPCT " and swap <= " PRIPCT "\n",
+    fprintf(stderr, "        SIGKILL when mem avail <= " PRIPCT " and swap free <= " PRIPCT "\n",
         args.mem_kill_percent, args.swap_kill_percent);
 
     startup_selftests(&args);
@@ -384,11 +440,11 @@ static unsigned sleep_time_ms(const poll_loop_args_t* args, const meminfo_t* m)
     const unsigned min_sleep = 100;
     const unsigned max_sleep = 1000;
 
-    long long mem_headroom_kib = (long long)((m->MemAvailablePercent - args->mem_term_percent) * 10 * (double)m->MemTotalMiB);
+    long long mem_headroom_kib = (long long)((m->MemAvailablePercent - args->mem_term_percent) * (double)m->UserMemTotalKiB / 100);
     if (mem_headroom_kib < 0) {
         mem_headroom_kib = 0;
     }
-    long long swap_headroom_kib = (long long)((m->SwapFreePercent - args->swap_term_percent) * 10 * (double)m->SwapTotalMiB);
+    long long swap_headroom_kib = (long long)((m->SwapFreePercent - args->swap_term_percent) * (double)m->SwapTotalKiB / 100);
     if (swap_headroom_kib < 0) {
         swap_headroom_kib = 0;
     }
@@ -451,7 +507,7 @@ static void poll_loop(const poll_loop_args_t* args)
                 kill_process(args, sig, &victim);
             }
         } else if (args->report_interval_ms && report_countdown_ms <= 0) {
-            print_mem_stats(printf, m);
+            print_mem_stats(info, m);
             report_countdown_ms = args->report_interval_ms;
         }
         unsigned sleep_ms = sleep_time_ms(args, &m);
