@@ -25,7 +25,7 @@ type cliTestCase struct {
 	stderrEmpty bool
 }
 
-func paseMeminfoLine(l string) int64 {
+func parseMeminfoLine(l string) int64 {
 	fields := strings.Split(l, " ")
 	asciiVal := fields[len(fields)-2]
 	val, err := strconv.ParseInt(asciiVal, 10, 64)
@@ -51,10 +51,10 @@ func parseMeminfo() (memTotal int64, swapTotal int64) {
 	lines := strings.Split(string(content), "\n")
 	for _, l := range lines {
 		if strings.HasPrefix(l, "MemTotal:") {
-			memTotal = paseMeminfoLine(l)
+			memTotal = parseMeminfoLine(l)
 		}
 		if strings.HasPrefix(l, "SwapTotal:") {
-			swapTotal = paseMeminfoLine(l)
+			swapTotal = parseMeminfoLine(l)
 		}
 	}
 	return
@@ -63,7 +63,8 @@ func parseMeminfo() (memTotal int64, swapTotal int64) {
 // earlyoom RSS should never be above 1 MiB,
 // but on some systems, it is (due to glibc?).
 // https://github.com/rfjakob/earlyoom/issues/221
-const rssMaxKiB = 2048
+// https://github.com/rfjakob/earlyoom/issues/296
+const rssMaxKiB = 4096
 
 func TestCli(t *testing.T) {
 	memTotal, swapTotal := parseMeminfo()
@@ -101,19 +102,21 @@ func TestCli(t *testing.T) {
 		// We use {"-r=0"} instead of {"-r", "0"} so runEarlyoom() can detect that there will be no output
 		{args: []string{"-r=0"}, code: -1, stderrContains: startupMsg, stdoutEmpty: true},
 		{args: []string{"-r", "0.1"}, code: -1, stderrContains: startupMsg, stdoutContains: memReport},
-		// Test --avoid and --prefer
+		// Test --avoid, --prefer, --ignore-root-user and --sort-by-rss
 		{args: []string{"--avoid", "MyProcess1"}, code: -1, stderrContains: "Will avoid killing", stdoutContains: memReport},
 		{args: []string{"--prefer", "MyProcess2"}, code: -1, stderrContains: "Preferring to kill", stdoutContains: memReport},
+		{args: []string{"--ignore-root-user"}, code: -1, stderrContains: "Processes owned by root will not be killed", stdoutContains: memReport},
+		{args: []string{"--sort-by-rss"}, code: -1, stderrContains: "Find process with the largest rss", stdoutContains: memReport},
 		{args: []string{"-i"}, code: -1, stderrContains: "Option -i is ignored"},
 		// Extra arguments should error out
 		{args: []string{"xyz"}, code: 13, stderrContains: "extra argument not understood", stdoutEmpty: true},
 		{args: []string{"-i", "1"}, code: 13, stderrContains: "extra argument not understood", stdoutEmpty: true},
 		// Tuples
-		{args: []string{"-m", "2,1"}, code: -1, stderrContains: "sending SIGTERM when mem <=  2.00% and swap <= 10.00%", stdoutContains: memReport},
+		{args: []string{"-m", "2,1"}, code: -1, stderrContains: "sending SIGTERM when mem avail <=  2.00% and swap free <= 10.00%", stdoutContains: memReport},
 		{args: []string{"-m", "1,2"}, code: -1, stdoutContains: memReport},
 		{args: []string{"-m", "1,-1"}, code: 15, stderrContains: "fatal", stdoutEmpty: true},
 		{args: []string{"-m", "1000,-1000"}, code: 15, stderrContains: "fatal", stdoutEmpty: true},
-		{args: []string{"-s", "2,1"}, code: -1, stderrContains: "sending SIGTERM when mem <= 10.00% and swap <=  2.00%", stdoutContains: memReport},
+		{args: []string{"-s", "2,1"}, code: -1, stderrContains: "sending SIGTERM when mem avail <= 10.00% and swap free <=  2.00%", stdoutContains: memReport},
 		{args: []string{"-s", "1,2"}, code: -1, stdoutContains: memReport},
 		// https://github.com/rfjakob/earlyoom/issues/97
 		{args: []string{"-m", "5,0"}, code: -1, stdoutContains: memReport},
@@ -132,18 +135,18 @@ func TestCli(t *testing.T) {
 		{args: []string{"-s", tooBigInt32}, code: 16, stderrContains: "fatal", stdoutEmpty: true},
 		{args: []string{"-s", tooBigUint32}, code: 16, stderrContains: "fatal", stdoutEmpty: true},
 		// Floating point values
-		{args: []string{"-m", "3.14"}, code: -1, stderrContains: "SIGTERM when mem <=  3.14%", stdoutContains: memReport},
-		{args: []string{"-m", "7,3.14"}, code: -1, stderrContains: "SIGKILL when mem <=  3.14%", stdoutContains: memReport},
-		{args: []string{"-s", "12.34"}, code: -1, stderrContains: "swap <= 12.34%", stdoutContains: memReport},
+		{args: []string{"-m", "3.14"}, code: -1, stderrContains: "SIGTERM when mem avail <=  3.14%", stdoutContains: memReport},
+		{args: []string{"-m", "7,3.14"}, code: -1, stderrContains: "SIGKILL when mem avail <=  3.14%", stdoutContains: memReport},
+		{args: []string{"-s", "12.34"}, code: -1, stderrContains: "swap free <= 12.34%", stdoutContains: memReport},
 		// Use both -m/-M
-		{args: []string{"-m", "10", "-M", mem1percent}, code: -1, stderrContains: "SIGTERM when mem <=  1.00%", stdoutContains: memReport},
+		{args: []string{"-m", "10", "-M", mem1percent}, code: -1, stderrContains: "SIGTERM when mem avail <=  1.00%", stdoutContains: memReport},
 	}
 	if swapTotal > 0 {
 		// Tests that cannot work when there is no swap enabled
 		tc := []cliTestCase{
 			{args: []string{"-S", swap2percent}, code: -1, stderrContains: " 2.00%", stdoutContains: memReport},
 			// Use both -s/-S
-			{args: []string{"-s", "10", "-S", swap2percent}, code: -1, stderrContains: "swap <=  1.00%", stdoutContains: memReport},
+			{args: []string{"-s", "10", "-S", swap2percent}, code: -1, stderrContains: "swap free <=  1.00%", stdoutContains: memReport},
 		}
 		testcases = append(testcases, tc...)
 	}
